"""The Python definition of the yarastorm Storm package."""


from stormlibpp import StormPkg

from . import __version__


SVC_NAME = "gormo.yara"
SVC_VER = __version__
SVC_GUID = "12bc0c8c1a37a29808d97972875d3913"
SVC_SYN_MIN_VER = (2, 137, 0)

SVC_EVTS = {
    "add": {
        "storm": f'[(meta:source={SVC_GUID} :name={SVC_NAME})]'
    }
}


class GormoYaraPkg(StormPkg):
    """The gormo.yara Storm package for the yarastorm service."""

    proto_name = SVC_NAME


PKGDEFS = (GormoYaraPkg().asdict(),)



"""Pytests for yarastorm."""


import binascii

import synapse.common as s_common
import synapse.tests.utils as s_tests
from yarastorm import __version__, YaraSvc
from yarastorm.defs import SVC_NAME


CMDS = (
    f"{SVC_NAME}.check",
    f"{SVC_NAME}.disable",
    f"{SVC_NAME}.enable",
    f"{SVC_NAME}.scan",
)
MODS = (f"{SVC_NAME}.lib",)


class TestYaraStorm(s_tests.SynTest):
    """Test suite for the yarastorm service."""

    async def test_svc_starts(self):
        async with self.getTestCoreProxSvc(YaraSvc) as (core, prox, svc):
            msgs = await core.stormlist("service.list")
            self.stormIsInPrint(f"true (svc) ({SVC_NAME} @ {__version__})", msgs)

    async def test_pkg_cmds_exist(self):
        async with self.getTestCoreProxSvc(YaraSvc) as (core, prox, svc):
            for cmd in CMDS:
                assert cmd in core.stormcmds

    async def test_pkg_mods_exist(self):
        async with self.getTestCoreProxSvc(YaraSvc) as (core, prox, svc):
            for mod in MODS:
                assert mod in core.stormmods

    async def test_pkg_mods_import(self):
        async with self.getTestCoreProxSvc(YaraSvc) as (core, prox, svc):
            for mod in MODS:
                msgs = await core.stormlist(f"$mod = $lib.import({mod})")
                self.stormHasNoErr(msgs)

    async def test_svc_import(self):
        async with self.getTestCoreProxSvc(YaraSvc) as (core, prox, svc):
            msgs = await core.stormlist(f"$svc = $lib.service.get({SVC_NAME})")
            self.stormHasNoErr(msgs)

    async def test_cmd_yara_check(self):
        async with self.getTestCoreProxSvc(YaraSvc) as (core, prox, svc):
            await core.nodes('[ it:app:yara:rule=* :text="test" :name="rule1"]')
            await core.nodes('[ it:app:yara:rule=* :text="rule dummy { condition: true }" :name="dummy" :enabled=$lib.true]')
            msgs = await core.stormlist(f"it:app:yara:rule | gormo.yara.check")
            print(msgs)
            self.stormHasNoErr(msgs)

    async def test_cmd_yara_scan(self):
        async with self.getTestAxon() as axon:
            _, sha256 = await axon.put(b"test")
            async with self.getTestCoreProxSvc(YaraSvc, ssvc_conf={"axon_url": axon.getLocalUrl()}) as (core, prox, svc):
                await core.nodes('[ it:app:yara:rule=* :text="test" :name="rule1"]')
                await core.nodes('[ it:app:yara:rule=* :text="rule dummy { condition: true }" :name="dummy" :enabled=$lib.true]')
                await core.nodes('[ file:bytes=* ]')
                await core.nodes(f'[ file:bytes={sha256.hex()} ]')
                msgs = await core.stormlist(f"file:bytes | gormo.yara.scan")
                print(msgs)
                self.stormHasNoErr(msgs)

    async def test_cmd_yara_enable(self):
        async with self.getTestCoreProxSvc(YaraSvc) as (core, prox, svc):
            rule_id = s_common.guid()
            await core.nodes(f'[ it:app:yara:rule={rule_id} :text="test" :name="rule1"]')
            msgs = await core.stormlist(f"it:app:yara:rule={rule_id} | gormo.yara.enable")
            print(msgs)
            self.stormHasNoErr(msgs)

    async def test_cmd_yara_disable(self):
        async with self.getTestCoreProxSvc(YaraSvc) as (core, prox, svc):
            rule_id = s_common.guid()
            await core.nodes(f'[ it:app:yara:rule={rule_id} :text="rule dummy {{ condition: true }}" :name="dummy" :enabled=$lib.true]')
            msgs = await core.stormlist(f"it:app:yara:rule={rule_id} | gormo.yara.disable")
            print(msgs)
            self.stormHasNoErr(msgs)



"""The CellApi implementation for yarastorm."""


from stormlibpp.telepath import BoolRetn
from stormlibpp.utils import normver
import synapse.lib.cell as s_cell
import synapse.lib.stormsvc as s_stormsvc

from .defs import SVC_EVTS, PKGDEFS, SVC_NAME, SVC_VER


class YaraApi(s_cell.CellApi, s_stormsvc.StormSvc):
    """The Telepath API endpoints for the triage-sandbox service."""

    # These defaults must be overridden from the StormSvc mixin
    _storm_svc_name = SVC_NAME
    _storm_svc_vers = normver(SVC_VER)[1]
    _storm_svc_evts = SVC_EVTS
    _storm_svc_pkgs = PKGDEFS

    async def matchFile(self, file_sha256: str, yara_rule) -> BoolRetn:
        """Test if a Yara rule matches a given file in the Axon."""

        return await self.cell.matchFile(file_sha256, yara_rule)

    async def compileRule(self, yara_rule, check: bool = False) -> BoolRetn:
        """Compile the given Yara rule and save it to this Cell's storage."""

        return await self.cell.compileRule(yara_rule, check)



"""The yarastorm service."""


import binascii
import os

from stormlibpp import utils
from stormlibpp.node import NodeTuple, StormNode
from stormlibpp.telepath import BoolRetn
import synapse.exc as s_exc
import synapse.lib.cell as s_cell
import synapse.telepath as s_telepath
import yara

from .api import YaraApi


class YaraSvc(s_cell.Cell):
    """The Cell implementation for the yarastorm service."""

    cellapi = YaraApi

    confdefs = {
        "axon_url": {
            "type": "string",
            "description": "The Telepath URL for an Axon service. "
            "This Axon is used to pull files for Yara matching.",
        },
        "rule_dir": {
            "type": "string",
            "description": "The directory that compiled Yara rules are saved in. "
            "This directory is relative to the Cell's 'dirn' path.",
            "default": "rules/",
        },
    }

    async def __anit__(self, dirn, *args, **kwargs):
        await s_cell.Cell.__anit__(self, dirn, *args, **kwargs)
        self.axonurl = self.conf.get("axon_url")
        self.ruledir = utils.absjoin(self.dirn, self.conf.get("rule_dir"))

    async def _getBytes(self, sha256: str) -> bytes | None:
        buffer = b""

        async with s_telepath.withTeleEnv():
            async with await s_telepath.openurl(self.axonurl) as axon:
                try:
                    async for part in axon.get(binascii.unhexlify(sha256)):
                        buffer += part
                except s_exc.NoSuchFile:
                    return None

        if len(buffer) == 0:
            return None

        return buffer

    async def matchFile(self, file_sha256: str, yara_rule: NodeTuple) -> BoolRetn:
        """Test if a Yara rule matches a given file in the Axon."""

        rulenode = StormNode.unpack(yara_rule)

        file_bytes = await self._getBytes(file_sha256)

        if file_bytes is None:
            return BoolRetn(
                status=False,
                mesg=f"Unable to find bytes for {file_sha256}",
                data=False
            )

        # TODO read the compiled yara file from disk if up to date
        rule = yara.compile(source=rulenode.props["text"], error_on_warning=True)

        if rule.match(data=file_bytes):
            return BoolRetn(status=True, mesg="", data=True)

        return BoolRetn(status=True, mesg="", data=False)

    async def compileRule(self, yara_rule: NodeTuple, check: bool = False) -> BoolRetn:
        """Compile the given Yara rule and save it to this Cell's storage."""

        rulenode = StormNode.unpack(yara_rule)

        try:
            rule = yara.compile(source=rulenode.props['text'], error_on_warning=True)
        except yara.SyntaxError as err:
            return BoolRetn(status=False, mesg=f"Yara Syntax Error - {err}", data=False)
        except yara.Error as err:
            return BoolRetn(status=False, mesg=f"Yara Error - {err}", data=False)

        if check:
            return BoolRetn(status=True, mesg="Successfully compiled rule!", data=True)

        # TODO actually save the compiled rule if check is false
        return BoolRetn(status=True, mesg="", data=True)
